#==============================================================================#
#
options("scipen" = 1000,digits = 18,"width" = 200)
est <- function(y,x0,x1,x2,val.inic){
  loglik <- function(theta,y,x0,x1,x2) {
    n <- length(y)
    a <- theta[1]
    g <- theta[2]
    l <- theta[3]
    s <- theta[4]
    (-n*log(s) - (n/2)*log(2*pi) -
        (1/(2*(s^2)))*sum( (y - a*x0 - g*x1 - l*x2)^2) )
  }

  grad.loglik <- function(theta, y,x0,x1,x2) {
    n <- length(y)
    a <- theta[1]
    g <- theta[2]
    l <- theta[3]
    s <- theta[4]
    g1 <- sum( (x0/(s^2))*(y - a*x0 - g*x1 - l*x2))
    g2 <- sum( (x1/(s^2))*(y - a*x0 - g*x1 - l*x2) )
    g3 <- sum( (x2/(s^2))*(y - a*x0 - g*x1 - l*x2) )
    g4 <- (-n/s) + (1/(s^3))*sum( (y - a*x0  - g*x1 - l*x2)^2 )
    c(g1,g2,g3,g4)
  }

  fit <- optim(par = val.inic, fn = loglik, gr = grad.loglik,
               method = "L-BFGS-B",
               lower = c(-Inf,-Inf,-Inf,0),
               upper = c( Inf, Inf,Inf,Inf),hessian = TRUE, #gama of -Inf to 0
               control = list(fnscale = -1,trace = 1,factr = 1e30,lmm = 5,maxit = 500),
               y = y,x0 = x0, x1 = x1, x2 = x2)


  
  return(fit)
}



#==============================================================================#
#

quad_change_point.LM <- function(data.set,var.indep,p1,p2){
#==============================================================================#
  
   data <- data.set
  
  a1 <- min(var.indep) 
  a2 <- max(var.indep) 
  
  
  
  psi1 <- p1 
  psi2 <- a2 
  #Variables
  data$alpha <- ifelse(data$Week <  psi2,1,0)
  data$U <- (ifelse(data$Week >= psi1 & data$Week < psi2,
                        data$Week - psi1, 0))^2
  #==============================================================================#
  #==============================================================================#
  # Fitting model zero

  X <- cbind(data$alpha,data$U)
  beta <- ginv(t(X) %*% X) %*% t(X) %*% data$Time 
  pred.fit <- X %*% beta
  
  # beta: alpha, beta e gama

  parameters <- as.numeric(beta)

  alpha <- parameters[1]

  gama <- parameters[2]
  #
  # Update lambda and psi
  lambda <- log( ( (psi1 - a1)/(a2 - psi1) ) )
  psi1 <- (a1 + a2*exp(lambda))/(1 + exp(lambda))
  

  epsilon <- 1

  param <- c(alpha,gama,lambda,1)
  init.v0 <-  as.matrix(c(alpha,gama,lambda,1))   
  iter <- 0
  
  # Step 3:
  # 
  while (abs(epsilon) > 1e-5 ) {
    # Matrices: hij, gij, G0 the ofsset and new yij
    hij <- -2*gama*(data$Week - psi1)*(as.numeric(I(data$Week >= psi1 & data$Week < psi2))) 
    gij <- (exp(lambda)*(a2 - a1))/((1 + exp(lambda))^2)
    data$V <- hij*gij
    data$G0 <- hij*gij*lambda
    data$TimeNew <- data$Time + data$G0
    
    data$alpha <- ifelse(data$Week <  psi2,1,0)
    data$U <- (ifelse(data$Week >= psi1 & data$Week < psi2,data$Week - psi1, 0))^2

    X <- cbind(data$alpha,data$U,data$V)

    
    fm <- try(est(data$TimeNew,data$alpha,data$U,data$V,param),TRUE)

    parameters <- c(fm$par) 
    pred.fit <- X[,1:2] %*% parameters[1:2]
    
    alpha <- as.numeric(parameters[1])
    gama <- as.numeric(parameters[2])
    #
    # Update lambda and psi
    lambda <- as.numeric(parameters[3])#log((psi1 - a1)/(a2 - psi1))
    psi1 <- (a1 + a2*exp(lambda))/(1 + exp(lambda))
    psi2 <- psi1 + sqrt(abs(-alpha/gama))
    psi.ind <-  cbind(psi1 = psi1,psi2 = psi2)
    
    init.v <- as.matrix(c(fm$par)) 
    epsilon <-  max((init.v - init.v0)/init.v0) 
    init.v0 <-   init.v 
    param <- fm$par

    cat( "\n")
    cat("iteration = ", iter <- iter + 1,'epsilon= ',epsilon,'psi1= ',psi1,"psi2=",psi2, "\n")
    if (iter >= 50000) { 
      break 
    }
    
  }
  
  #==============================================================================#
  X.p <- cbind(data$alpha,data$U,data$V)
  pred_indf <- X.p %*% parameters[1:3] 
  mspe <- mean(((data$TimeNew - pred_indf)^2))
  var.par <- diag(ginv(-fm$hessian[1:3,1:3]))
  values <- list(iter = iter,estimates = parameters[1:3],psi.ind = psi.ind,Std.Error = var.par,mspe = mspe,fitted.ind = pred_indf,
                 vcov.beta = ginv(-fm$hessian[1:3,1:3]))
  return(values)
}


