###################################################################################
# R function to fit segmented polynomial mixed models according to:
# Singer et al. (2019)
# This function is required by the R code "Fittingcode.R"
## Authors: 
#################################################################################


#######################################################################################
# Package
##
library(nlme)

#==============================================================================#

quad_change_point <- function(data.set,var.indep,p1,p2){
#==============================================================================#

  vecdados <- data.set
  
  a1 <- min(var.indep) 
  a2 <- max(var.indep) 
  
  psi1 <- p1 
  psi2 <- p2 

  vecdados$alpha <- ifelse(vecdados$Week <  psi2,1,0)
  vecdados$U <- (ifelse(vecdados$Week >= psi1 & vecdados$Week < psi2,
                        vecdados$Week - psi1, 0))^2

    Ctrl <- lmeControl( msVerbose = T,
                        opt = "optim",
                        optimMethod = 'BFGS',
                        niterEM = 20000) #20000 #Perycite_M em 10000 pericyteF 1000

#==============================================================================#
  # Fitting model zero
  
  fit <- lme(Time ~ alpha + U - 1 , 
             random = ~ alpha + U - 1|ind, data = vecdados,
             control = Ctrl)
  summary(fit)
  rand.coefpred <- (fit$coefficients)

  parameters <- as.numeric(rand.coefpred$fixed)

  alpha <- parameters[1]
  gama <- parameters[2]

  alphai <- parameters[1] + ranef(fit)[,1]  
  Alphai <- rep( alphai,times = as.numeric(table(fit$groups)) ) 
 
  gamai <- parameters[2] + ranef(fit)[,2]
  Gamai <- rep( gamai,times = as.numeric(table(fit$groups)) ) 
  
  # Update lambda and psi
  lambda <- log((psi1 - a1)/(a2 - psi1))
  psi1 <- (a1 + a2*exp(lambda))/(1 + exp(lambda))
  psi2i <- psi2                                     #psi1 + sqrt(abs(Alphai/Gamai))
  # Matrices: hij, gij, G0 the ofsset and new yij
  hij <- -2*Gamai*(vecdados$Week - psi1)*(as.numeric(I(vecdados$Week >= psi1 & vecdados$Week < psi2i))) 
  gij <- (exp(lambda)*(a2 - a1))/((1 + exp(lambda))^2)
  vecdados$V <- hij*gij
  vecdados$G0 <- hij*gij*lambda
  vecdados$TimeNew <- vecdados$Time + vecdados$G0
  
  vecdados$alpha <- ifelse(vecdados$Week <  psi2i,1,0)
  vecdados$U <- (ifelse(vecdados$Week >= psi1 & vecdados$Week < psi2i,vecdados$Week - psi1, 0))^2
 
  # Group data
  vecdados <- groupedData(Time ~  U  |ind, data = vecdados)
  
  epsilon <- 1
  LogLik0 <- abs(fit$logLik)
  iter <- 0
  
  # Step 3:
  while (abs(epsilon) > 1e-3 ) {
    
    fit <- lme(TimeNew ~ alpha + U + V - 1,
               random = ~ alpha + U + V - 1|id,
               control = Ctrl,method = 'REML',data = vecdados)
    summary(fit)
    rand.coefpred <- (fit$coefficients)
    
    # New values for Lambda and psi
    lambda <- as.numeric(rand.coefpred$fixed[3])
    lambdai <- lambda + ranef(fit)[,3]
    Lambdai <- rep(lambdai, times = as.numeric(table(fit$groups)) )
    psi1 <- (a1 + a2*exp( lambda ))/(1 + exp(lambda))
    psi1i <- (a1 + a2*exp( lambdai ))/(1 + exp(lambdai))
    Psi1i <- rep(psi1i, times = as.numeric(table(fit$groups)) )
    
    parameters <- as.numeric(rand.coefpred$fixed)
    alpha <- parameters[1]
    gama <- parameters[2]
    alphai <- parameters[1] + ranef(fit)[,1]  
    Alphai <- rep( alphai,times = as.numeric(table(fit$groups)) ) 
    gamai <- parameters[2] + ranef(fit)[,2]
    Gamai <- rep( gamai,times = as.numeric(table(fit$groups)) ) 
    
    psi2 <- psi1 + sqrt(abs(-alpha/gama))
    psi2i <- psi1i + sqrt(abs(-alphai/gamai))
    Psi2i <- rep(psi2i, times = as.numeric(table(fit$groups)) )

    vecdados$alpha <- ifelse(vecdados$Week <  Psi2i,1,0)
    vecdados$U <- (ifelse(vecdados$Week >= Psi1i & vecdados$Week < Psi2i,vecdados$Week - Psi1i, 0))^2
    
    # Matrix hij, g, offset
    hij <- -2*Gamai*(vecdados$Week - Psi1i)*(as.numeric(I(vecdados$Week >= Psi1i & vecdados$Week < Psi2i))) 
    gij <- (exp(Lambdai)*(a2 - a1))/((1 + exp(Lambdai))^2)
    vecdados$V <- hij*gij
    vecdados$G0 <- hij*gij*Lambdai 
   
    vecdados$TimeNew <- vecdados$Time + vecdados$G0
    
    psi.ti <-  cbind(psi1 = psi1,psi2 = psi2)
    psi.ind <-  cbind(psi1i,psi2i)
    pred.par.ind <- cbind(alphai = alphai,gamai = gamai,lambdai = lambdai)
    LogLik <- abs(fit$logLik)
    epsilon <- abs((LogLik - LogLik0) / LogLik0) 
    LogLik0 <- LogLik
    cat( "\n")
    cat("iteration = ", iter <- iter + 1,'epsilon= ',epsilon,'psi1= ',psi1,"psi2=",psi2, "\n")
    if (iter >= 500) {
      break 
    }
    
  }
  
#==============================================================================#
  #  return(values.mean)
  pred_ind <- as.numeric(predict(fit,newdata = vecdados,level = 1)) 
  mspe <- as.numeric(mean(((as.numeric(vecdados$TimeNew) - pred_ind)^2)))
  values <- list(iter = iter,fm = fit,psi.ti = psi.ti,
                 psi.ind = psi.ind,mspe = mspe,
                 pred.par.ind = pred.par.ind)
  return(values)
}



